/*
  License text attribution: zlib license at

  http://www.gzip.org/zlib/zlib_license.html

  This license is a modification of the zlib license, changing names and
  dates.

  Copyright (C) 2007 Samson Chee

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Samson Chee samson@herethere.net
*/

#include "stdafx.h"

#include <stdio.h>
#include "bubble_data.h"

tBDBool isOneColor(tBDColor pColor)
{
  return
    (pColor == (1 << 0)) ||
    (pColor == (1 << 1)) ||
    (pColor == (1 << 2)) ||
    (pColor == (1 << 3)) ||
    (pColor == (1 << 4)) ||
    (pColor == (1 << 5)) ||
    (pColor == (1 << 6)) ||
    (pColor == (1 << 7));
}

tBDColor getRandColor(sBubbleMode* pBMode)
{
  return (1 << (rand() % pBMode->mColors));
}

int groupScore(sBubbleMode* pBMode, tBDDim2 pSize)
{
  // scoring formula attribution: http://en.wikipedia.org/wiki/SameGame
  if (pSize <= 1)
  {
    return 0;
  }

  switch (pBMode->mScoreMode)
  {
  case 1 :
    {
      // straight scoring (unique)
      return pSize;
    }; break;
  case 2 :
    {
      // summation formula for i += 2..n (unique)
      return ((pSize * (pSize + 1)) / 2) - 1;
    }; break;
  case 3 :
    {
      // SameGame original (variant 1 - Ikuo Hirohata's)
      return ((pSize - 2) * (pSize - 2));
    }; break;
  case 4 :
    {
      // SameGame original (variant 2 - insame game / TI)
      return ((pSize - 1) * (pSize - 1));
    }; break;
  case 5 :
    {
      // Bubble Breaker
      return (pSize * (pSize - 1));
    }; break;
  case 6 :
    {
      // Jawbreaker (http://en.wikipedia.org/wiki/Jawbreaker_(Pocket_PC_game))
      return ((pSize * pSize) - pSize);
    }; break;
  }

  return 0;
}

sBubbleData* bd_init()
{
  // init 2d matrix and also both 1d queue matrixes
  // - NOTE: the allocated memory is NOT off by one because we store our queues
  //   in the space of the last node
  tBDDim2 theNumElements = (kBubbleDataMaxX * kBubbleDataMaxY) + kBubbleDataMaxX + kBubbleDataMaxY;
  unsigned int theMemSize = sizeof(sBubbleData) + (theNumElements * sizeof(sBubbleNode));

  sBubbleData* theData = (sBubbleData*)malloc(theMemSize);
  theData->mMemSize = theMemSize;

  return theData;
}

sBubbleData* bd_copy(sBubbleData* pBDTarget, sBubbleData* pBDSource)
{
  if (pBDTarget->mMemSize != pBDSource->mMemSize)
  {
    return 0;
  }

  memcpy(pBDTarget, pBDSource, pBDSource->mMemSize);

  return pBDTarget;
}

void bd_destroy(sBubbleData* pBD)
{
  free(pBD);
}

tBDBool bd_game_active(sBubbleData* pBD)
{
  // return true if there is at least one pair (which means that the game is not over yet)
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      tBDBool theOk = (bd_mark_select(pBD, i, j) > 1);
      bd_reset_hit_mark_move(pBD);

      if (theOk)
      {        
        return 1;
      }
    }
  }

  return 0;
}

void bd_new_game(sBubbleMode* pBMode, sBubbleData* pBD)
{
  bd_reset(pBD);

  bd_randomize_color(pBMode, pBD);
  while (!bd_game_active(pBD))
  {
    bd_randomize_color(pBMode, pBD);
  }

  bd_randomize_color_qx(pBMode, pBD);
  bd_randomize_color_qy(pBMode, pBD);

  pBD->mNumSelected = 0;
  pBD->mLargestGroup = 0;
  pBD->mMaxMove = 0;
  pBD->mMoves = 0;
  pBD->mScore = 0;

  pBD->mLastGroupSize = 0;
  pBD->mLastGroupScore = 0;
}

sBubbleNode* bd_ptr_1d(sBubbleData* pBD, tBDDim2 pNum)
{
  if (pNum >= kBubbleDataMaxX * kBubbleDataMaxY)
  {
    return 0;
  }

  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + pNum;
}

sBubbleNode* bd_ptr(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  if ((pX >= kBubbleDataMaxX) || (pY >= kBubbleDataMaxY))
  {
    return 0;
  }

  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + ((pY * kBubbleDataMaxX) + pX);
}

void bd_reset(sBubbleData* pBD)
{
  memset(pBD->mNodes, 0, kBubbleDataMaxX * kBubbleDataMaxY * sizeof(sBubbleNode));
}

void bd_reset_hit(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mHit = 0;
  }
}

void bd_randomize_color(sBubbleMode* pBMode, sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mColor = getRandColor(pBMode);
  }
}

void bd_reset_mark(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mMark = 0;
  }

  pBD->mNumSelected = 0;
}

void bd_reset_move(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    bd_ptr_1d(pBD, i)->mMove = 0;
  }
}

void bd_reset_hit_mark_move(sBubbleData* pBD)
{
  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    sBubbleNode* theNode = bd_ptr_1d(pBD, i);
    
    theNode->mHit = 0;
    theNode->mMark = 0;
    theNode->mMove = 0;
  }

  pBD->mNumSelected = 0;
}

void bd_mark(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  bd_ptr(pBD, pX, pY)->mMark = 1;
}

tBDBool bd_ismarked(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  return bd_ptr(pBD, pX, pY)->mMark;
}

tBDDim2 bd_mark_select(sBubbleData* pBD, tBDDim pX, tBDDim pY)
{
  bd_reset_mark(pBD);
  bd_reset_hit(pBD);
  pBD->mNumSelected = bd_mark_select(pBD, pX, pY, 0xff);

  return pBD->mNumSelected;
}

tBDDim2 bd_mark_select(sBubbleData* pBD, tBDDim pX, tBDDim pY, tBDColor pColor)
{
  sBubbleNode* theNode = bd_ptr(pBD, pX, pY);

  // ignore marked or hit nodes because we've already been here
  if ((theNode->mColor == 0) || (theNode->mHit & theNode->mColor) || (theNode->mMark) || (theNode->mMove))
  {
    return 0;
  }

  // node is not marked, but it may already have been hit by another color
  // - as long as we remember to check the bit flags, we'll be fine

  // hit the node if the colors don't match
  if ((theNode->mColor & pColor) == 0)
  {
    theNode->mHit |= pColor;
    return 0;
  }

  // mark this node
  theNode->mMark = 1;

  // go look at neighbouring nodes
  tBDDim2 theAcc = 1;
  tBDColor theNodeColor = isOneColor(theNode->mColor) ? theNode->mColor : pColor;

  if (pX > 0)
  {
    theAcc += bd_mark_select(pBD, pX - 1, pY, theNodeColor);
  }

  if (pX < kBubbleDataMaxX - 1)
  {
    theAcc += bd_mark_select(pBD, pX + 1, pY, theNodeColor);
  }

  if (pY > 0)
  {
    theAcc += bd_mark_select(pBD, pX, pY - 1, theNodeColor);
  }

  if (pY < kBubbleDataMaxY - 1)
  {
    theAcc += bd_mark_select(pBD, pX, pY + 1, theNodeColor);
  }

  return theAcc;
}

tBDDim2 bd_mark_clear(sBubbleMode* pBMode, sBubbleData* pBD)
{
  tBDDim2 theAcc = 0;

  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    sBubbleNode* theBD = bd_ptr_1d(pBD, i);

    // clear only the marked nodes
    if (theBD->mMark)
    {
      theBD->mColor = 0;
      theAcc++;
    }

    theBD->mHit = 0;
    theBD->mMark = 0;
    theBD->mMove = 0;
  }

  pBD->mNumSelected = 0;
  pBD->mLargestGroup = (pBD->mLargestGroup > theAcc) ? pBD->mLargestGroup : theAcc;
  pBD->mMoves++;

  pBD->mLastGroupSize = theAcc;
  pBD->mLastGroupScore = groupScore(pBMode, theAcc);

  pBD->mScore += pBD->mLastGroupScore;

  return theAcc;
}

tBDDim2 bd_convert_move_to_mark(sBubbleData* pBD)
{
  pBD->mNumSelected = 0;

  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      if (theNode->mMove > 1)
      {
        theNode->mMark = 1;
        pBD->mNumSelected++;
      }
      else
      {
        theNode->mMark = 0;
      }

      theNode->mMove = 0;
    }
  }

  return pBD->mNumSelected;
}

tBDDim2 bd_convert_mark_to_move(sBubbleData* pBD, tBDDim pMove)
{
  tBDDim2 theAcc = 0;

  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      if (theNode->mMark)
      {
        theNode->mMark = 0;
        theNode->mMove = pMove;

        theAcc++;
      }

    }
  }

  return theAcc;
}

tBDDim2 bd_count_by_color(sBubbleData* pBD, tBDColor pColor)
{
  tBDDim2 theAcc = 0;

  for (tBDDim2 i = 0; i < kBubbleDataMaxX * kBubbleDataMaxY; i++)
  {
    sBubbleNode* theBD = bd_ptr_1d(pBD, i);
    theAcc += (theBD->mColor == pColor) ? 1 : 0;
  }

  return theAcc;
}

tBDDim2 bd_count_by_color_min_connected(sBubbleData* pBD, tBDColor pColor, tBDDim2 pMinSelected)
{
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      tBDDim2 theMarked = bd_mark_select(pBD, i, j, pColor);

      if (theMarked)
      {
        tBDDim theMoveFlag = (theMarked >= pMinSelected) ? 2 : 1;

        if (theMarked == 1)
        {
          sBubbleNode* theNode = bd_ptr(pBD, i, j);
          theNode->mMark = 0;
          theNode->mHit = 0;
          theNode->mMove = theMoveFlag;
        }
        else
        {
          bd_convert_mark_to_move(pBD, theMoveFlag);
        }
      }

    }
  }

  tBDDim2 theMoveFlagged = bd_convert_move_to_mark(pBD);
  bd_reset_hit_mark_move(pBD);

  return theMoveFlagged;
}

tBDDim2 bd_count_unconnected(sBubbleData* pBD)
{
  tBDDim2 theAcc = 0;

  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      if (theNode->mColor == 0)
      {
        continue;
      }

      if (i > 0)
      {
        sBubbleNode* theTempNode = bd_ptr(pBD, i - 1, j);
        if (theTempNode->mColor & theNode->mColor)
        {
          continue;
        }
      }

      if (i < kBubbleDataMaxX - 1)
      {
        sBubbleNode* theTempNode = bd_ptr(pBD, i + 1, j);
        if (theTempNode->mColor & theNode->mColor)
        {
          continue;
        }
      }

      if (j > 0)
      {
        sBubbleNode* theTempNode = bd_ptr(pBD, i, j - 1);
        if (theTempNode->mColor & theNode->mColor)
        {
          continue;
        }
      }

      if (j < kBubbleDataMaxY - 1)
      {
        sBubbleNode* theTempNode = bd_ptr(pBD, i, j + 1);
        if (theTempNode->mColor & theNode->mColor)
        {
          continue;
        }
      }

      theAcc++;
    }
  }

  return theAcc;
}

int bd_count_delta_unconnected_for_selected(sBubbleMode* pBMode, sBubbleData* pBD, sBubbleData* pTempBD, bool pCheckQueues)
{
  bd_copy(pTempBD, pBD);

  bd_reset_move(pTempBD);

  tBDDim2 theAccPre = 0;
  tBDDim2 theAccPost = 0;

  theAccPre += bd_count_unconnected(pTempBD);
  theAccPre += kBubbleDataMaxX + kBubbleDataMaxY;

  bd_mark_clear(pBMode, pTempBD);

  bd_gravityY_pre(pBMode, pTempBD);
  bd_gravityY_post(pBMode, pTempBD);

  if (pCheckQueues && bd_can_load_qx(pBMode, pTempBD))
  {
    bd_load_qx(pBMode, pTempBD, 0);

    bd_gravityY_pre(pBMode, pTempBD);
    bd_gravityY_post(pBMode, pTempBD);
  }
  else
  {
    theAccPost += kBubbleDataMaxX;
  }

  bd_gravityX_pre(pBMode, pTempBD);
  bd_gravityX_post(pBMode, pTempBD);

  if (pCheckQueues && bd_can_load_qy(pBMode, pTempBD))
  {
    bd_load_qy(pBMode, pTempBD, 0);

    bd_gravityX_pre(pBMode, pTempBD);
    bd_gravityX_post(pBMode, pTempBD);
  }
  else
  {
    theAccPost += kBubbleDataMaxY;
  }

  theAccPost += bd_count_unconnected(pTempBD);

  return theAccPost - theAccPre;
}

tBDColor bd_get_max_count_color(sBubbleData* pBD, tBDDim2 pMinSelected)
{
  tBDDim2 c1, c2, c3, c4;

  if (pMinSelected == 0)
  {
    c1 = bd_count_by_color(pBD, 0x01);
    c2 = bd_count_by_color(pBD, 0x02);
    c3 = bd_count_by_color(pBD, 0x04);
    c4 = bd_count_by_color(pBD, 0x08);
  }
  else
  {
    c1 = bd_count_by_color_min_connected(pBD, 0x01, pMinSelected);
    c2 = bd_count_by_color_min_connected(pBD, 0x02, pMinSelected);
    c3 = bd_count_by_color_min_connected(pBD, 0x04, pMinSelected);
    c4 = bd_count_by_color_min_connected(pBD, 0x08, pMinSelected);
  }

  if ((c1 > c2) && (c1 > c3) && (c1 > c4))
  {
    return 0x01;
  }
  else if ((c2 > c3) && (c2 > c4))
  {
    return 0x02;
  }
  else if (c3 > c4)
  {
    return 0x04;
  }
  else if (c4)
  {
    return 0x08;
  }

  return 0;
}

tBDDim2 bd_mark_clearable_column_group(sBubbleMode* pBMode, sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    tBDColor theColor = 0;
    tBDDim theCount = 0;

    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      if (theNode->mColor)
      {
        if (j == 0)
        {
          theColor = theNode->mColor;
          theCount++;
        }
        else
        {
          if (theColor == theNode->mColor)
          {
            theCount++;
          }
          else
          {
            theCount = 0;
            break;
          }
        }
      }
      else
      {
        break;
      }
    }

    if (theCount)
    {
      tBDDim2 theNumSelected = bd_mark_select(pBD, i, 0);

      if (theNumSelected > 1)
      {
        return theNumSelected;
      }
      else
      {
        sBubbleNode* theNode = bd_ptr(pBD, i, j);
        theNode->mMark = 0;
      }
    }
  }

  return 0;
}

tBDDim2 bd_mark_highest_row_group(sBubbleMode* pBMode, sBubbleData* pBD, tBDDim2 pMinSelected, tBDColor pExcludeColor)
{
  // can't do the j for-loop backwards because the >= 0 test will fail for unsigned short
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    tBDDim theRow = (kBubbleDataMaxY - 1) - j;
    tBDColor theColor = 0;
    int theCount = 0;

    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, theRow);

      if (theColor && (theNode->mColor == theColor) && (theNode->mColor != pExcludeColor))
      {
        tBDDim2 theNumSelected = bd_mark_select(pBD, i, theRow);
        if (theNumSelected < pMinSelected)
        {
          bd_reset_mark(pBD);
        }
        else
        {
          return pBD->mNumSelected;
        }
      }

      theColor = theNode->mColor;
    }
  }

  return 0;
}

tBDDim2 bd_mark_lowest_post_unconnected(sBubbleMode* pBMode, sBubbleData* pBD, sBubbleData* pTempBD, tBDColor pExcludeColor, bool pCheckQueues)
{
  tBDDim theFoundX = 0;
  tBDDim theFoundY = 0;
  int theMaxDelta = kBubbleDataMaxX * kBubbleDataMaxY;

  for (tBDDim a = 0; a < (kBubbleDataMaxX + kBubbleDataMaxY) - 1; a++)
  {
    // check index
    tBDDim theMaxX = (a >= kBubbleDataMaxX) ? kBubbleDataMaxX - 1: a;

    // we want to count down, but we can't go past -1, so theX inverts i
    for (tBDDim i = 0; i <= theMaxX; i++)
    {
      tBDDim theX = theMaxX - i;
      tBDDim j = a - theX;

      if (j >= kBubbleDataMaxY)
      {
        break;
      }

      sBubbleNode* theNode = bd_ptr(pBD, theX, j);
      if (theNode->mColor & pExcludeColor)
      {
        continue;
      }

      tBDDim2 theNumSelected = bd_mark_select(pBD, theX, j);
      if (theNumSelected <= 1)
      {
        theNode->mMove = 1;
        continue;
      }

      int theDelta = bd_count_delta_unconnected_for_selected(pBMode, pBD, pTempBD, pCheckQueues);
      bd_convert_mark_to_move(pBD, theX);

      if (theDelta < theMaxDelta)
      {
        theFoundX = theX;
        theFoundY = j;
        theMaxDelta = theDelta;
      }

    }
  }

  bd_reset_hit_mark_move(pBD);
  bd_mark_select(pBD, theFoundX, theFoundY);

  if (pBD->mNumSelected == 1)
  {
    sBubbleNode* theNode = bd_ptr(pBD, theFoundX, theFoundY);

    theNode->mHit = 0;
    theNode->mMark = 0;
    theNode->mMove = 0;

    pBD->mNumSelected = 0;
  }

  return pBD->mNumSelected;
}

tBDDim2 bd_mark_largest_group(sBubbleMode* pBMode, sBubbleData* pBD)
{
  // get the largest group by looking at each node
  // - mark each group that we finish looking at using the mMove member variable as temporary storage
  // - groups of 1 are always marked as 1
  // - groups smaller than the largest group are also marked as 1

  tBDDim2 theLargestGroup = 0;

  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      if ((theNode->mColor == 0) || (theNode->mMove))
      {
        continue;
      }

      tBDDim2 theGroupSize = bd_mark_select(pBD, i, j);

      if (theGroupSize <= 1)
      {
        continue;
      }

      for (tBDDim x = 0; x < kBubbleDataMaxX; x++)
      {
        for (tBDDim y = 0; y < kBubbleDataMaxY; y++)
        {
          sBubbleNode* theNode = bd_ptr(pBD, x, y);

          if (theNode->mColor == 0)
          {
            continue;
          }

          if (theGroupSize > theLargestGroup)
          {
            if (theNode->mMark)
            {
              theNode->mMove = 2;
            }
            else
            {
              if (theNode->mMove > 1)
              {
                theNode->mMove = 1;
              }
            }
          }
          else
          {
            theNode->mMove = (theNode->mMark) ? 1 : theNode->mMove;
          }            
        }
      }

      theLargestGroup = (theGroupSize > theLargestGroup) ? theGroupSize : theLargestGroup;
    }
  }

  return bd_convert_move_to_mark(pBD);
}

tBDDim2 bd_mark_hint_group(sBubbleMode* pBMode, sBubbleData* pBD, int pMode, sBubbleData* pTempBD)
{
  // mode 1 is the simplest mode and just selects the largest group
  if (pMode == 1)
  {
    // largest group is default strategy
    if (bd_mark_largest_group(pBMode, pBD))
    {
      return pBD->mNumSelected;
    }

    return 0;
  }

  // modes > 1 are smart enough to use the bd_mark_clearable_column_group
  // strategy first
  if (bd_mark_clearable_column_group(pBMode, pBD))
  {
    return pBD->mNumSelected;
  }

  switch (pMode)
  {
  case 2:
    {
      if (bd_mark_highest_row_group(pBMode, pBD, 3, 0))
      {
        return pBD->mNumSelected;
      }
    }; break;
  case 3:
    {
      if (bd_mark_highest_row_group(pBMode, pBD, 3, bd_get_max_count_color(pBD, 2)))
      {
        return pBD->mNumSelected;
      }
    }; break;
  case 4:
    {
      if (bd_mark_lowest_post_unconnected(pBMode, pBD, pTempBD, 0, false))
      {
        return pBD->mNumSelected;
      }
    }; break;
  case 5:
    {
      if (bd_mark_lowest_post_unconnected(pBMode, pBD, pTempBD, 0, true))
      {
        return pBD->mNumSelected;
      }
    }; break;
  case 6:
    {
      if (bd_mark_lowest_post_unconnected(pBMode, pBD, pTempBD, bd_get_max_count_color(pBD, 0), false))
      {
        return pBD->mNumSelected;
      }
    }; break;
  case 7:
    {
      if (bd_mark_lowest_post_unconnected(pBMode, pBD, pTempBD, bd_get_max_count_color(pBD, 0), true))
      {
        return pBD->mNumSelected;
      }
    }; break;
  }

  // largest group is default strategy
  if (bd_mark_largest_group(pBMode, pBD))
  {
    return pBD->mNumSelected;
  }

  return 0;
}

void bd_move_node(sBubbleData* pBD, tBDDim pOrgX, tBDDim pOrgY, tBDDim pNewX, tBDDim pNewY)
{
  sBubbleNode* theOrgNode = bd_ptr(pBD, pOrgX, pOrgY);
  sBubbleNode* theNewNode = bd_ptr(pBD, pNewX, pNewY);

  theNewNode->mColor = theOrgNode->mColor;
  theOrgNode->mColor = 0;

  theNewNode->mMove = 0;
  theOrgNode->mMove = 0;
}

void bd_gravityX_pre(sBubbleMode* pBMode, sBubbleData* pBD)
{
  // calculate how much to drop
  if (pBMode->mFullGravityX)
  {
    // shift everything left to right - like bd_gravityY_pre
    tBDDim theMoveAcc = 0;
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      theMoveAcc = 0;
      for (tBDDim i = 0; i < kBubbleDataMaxY; i++)
      {
        sBubbleNode* theNode = bd_ptr(pBD, i, j);
        if (theNode->mColor)
        {
          theNode->mMove = theMoveAcc;
          pBD->mMaxMove = (pBD->mMaxMove > theMoveAcc) ? pBD->mMaxMove : theMoveAcc;
        }
        else
        {
          theMoveAcc++;
        }
      }
    }
  }
  else
  {
    // shift only to fill voids left by empty columns
    tBDDim theMoveAcc = 0;
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      tBDBool theColEmpty = 1;
      for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
      {
        sBubbleNode* theNode = bd_ptr(pBD, i, j);
        if (theNode->mColor)
        {
          theNode->mMove = theMoveAcc;
          theColEmpty = 0;

          pBD->mMaxMove = (pBD->mMaxMove > theMoveAcc) ? pBD->mMaxMove : theMoveAcc;
        }
      }

      if (theColEmpty)
      {
        theMoveAcc++;
      }
    }
  }
}

void bd_gravityX_post(sBubbleMode* pBMode, sBubbleData* pBD)
{
  // drop everything down
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);
      if (theNode->mMove)
      {
        // move to last hole and update last hole to the position above
        bd_move_node(pBD, i, j, i - theNode->mMove, j);
      }
    }
  }

  pBD->mMaxMove = 0;
}

void bd_gravityY_pre(sBubbleMode* pBMode, sBubbleData* pBD)
{
  // calculate how much to drop
  tBDDim theMoveAcc = 0;
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    theMoveAcc = 0;
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);
      if (theNode->mColor)
      {
        theNode->mMove = theMoveAcc;
        pBD->mMaxMove = (pBD->mMaxMove > theMoveAcc) ? pBD->mMaxMove : theMoveAcc;
      }
      else
      {
        theMoveAcc++;
      }
    }
  }
}

void bd_gravityY_post(sBubbleMode* pBMode, sBubbleData* pBD)
{
  // drop everything down
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);
      if (theNode->mMove)
      {
        // move to last hole and update last hole to the position above
        bd_move_node(pBD, i, j, i, j - theNode->mMove);
      }
    }
  }

  pBD->mMaxMove = 0;
}

sBubbleNode* bd_ptr_qx(sBubbleData* pBD, tBDDim pNum)
{
  if (pNum >= kBubbleDataMaxX)
  {
    return 0;
  }

  tBDDim2 theOffset = (kBubbleDataMaxX * kBubbleDataMaxY) + pNum;
  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + theOffset;
}

sBubbleNode* bd_ptr_qy(sBubbleData* pBD, tBDDim pNum)
{
  if (pNum >= kBubbleDataMaxY)
  {
    return 0;
  }

  tBDDim2 theOffset = (kBubbleDataMaxX * kBubbleDataMaxY) + kBubbleDataMaxX + pNum;
  // note: no sizeof operator required here because pBD is properly typed already
  sBubbleNode* theBN = pBD->mNodes;
  return theBN + theOffset;
}

void bd_randomize_color_qx(sBubbleMode* pBMode, sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theNode = bd_ptr_qx(pBD, i);
    theNode->mColor = (pBMode->mUseQx) ? getRandColor(pBMode) : 0;
  }
}

void bd_randomize_color_qy(sBubbleMode* pBMode, sBubbleData* pBD)
{
  for (tBDDim i = 0; i < kBubbleDataMaxY; i++)
  {
    sBubbleNode* theNode = bd_ptr_qy(pBD, i);
    theNode->mColor = (pBMode->mUseQy) ? getRandColor(pBMode) : 0;
  }
}

tBDBool bd_can_load_qx(sBubbleMode* pBMode, sBubbleData* pBD)
{
  if (!pBMode->mUseQx)
  {
    return 0;
  }

  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theNode = bd_ptr(pBD, i, kBubbleDataMaxY - 1);
    if (theNode->mColor)
    {
      return 0;
    }
  }

  return 1;
}

tBDBool bd_can_load_qy(sBubbleMode* pBMode, sBubbleData* pBD)
{
  if (!pBMode->mUseQy)
  {
    return 0;
  }

  for (tBDDim i = 0; i < kBubbleDataMaxY; i++)
  {
    sBubbleNode* theNode = bd_ptr(pBD, kBubbleDataMaxX - 1, i);
    if (theNode->mColor)
    {
      return 0;
    }
  }

  return 1;
}

void bd_load_qx(sBubbleMode* pBMode, sBubbleData* pBD, bool pRandomizeNext)
{
  if (!pBMode->mUseQx)
  {
    return;
  }

  for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
  {
    sBubbleNode* theMatrixNode = bd_ptr(pBD, i, kBubbleDataMaxY - 1);
    sBubbleNode* theQueueNode = bd_ptr_qx(pBD, i);

    theMatrixNode->mColor = theQueueNode->mColor;
  }

  if (pRandomizeNext)
  {
    bd_randomize_color_qx(pBMode, pBD);
  }
}

void bd_load_qy(sBubbleMode* pBMode, sBubbleData* pBD, bool pRandomizeNext)
{
  if (!pBMode->mUseQy)
  {
    return;
  }

  for (tBDDim i = 0; i < kBubbleDataMaxY; i++)
  {
    sBubbleNode* theMatrixNode = bd_ptr(pBD, kBubbleDataMaxX - 1, i);
    sBubbleNode* theQueueNode = bd_ptr_qy(pBD, i);

    theMatrixNode->mColor = theQueueNode->mColor;
  }

  if (pRandomizeNext)
  {
    bd_randomize_color_qy(pBMode, pBD);
  }
}

void bd_debug_output_matrix(sBubbleData* pBD)
{
#ifdef WIN32
  // print color values
  for (tBDDim j = 0; j < kBubbleDataMaxY; j++)
  {
    for (tBDDim i = 0; i < kBubbleDataMaxX; i++)
    {
      sBubbleNode* theNode = bd_ptr(pBD, i, j);

      char theBuf[6];
      sprintf(theBuf, "%02X%c%c ", theNode->mColor, theNode->mMove ? (theNode->mMove > 1 ? '*' : '+') : '.', theNode->mMark ? '+' : '.');
      OutputDebugString(theBuf);
    }

    OutputDebugString("\n");
  }
#endif
}
